@@ -27,7 +27,7 @@
 #include "x_keymap.h"
 
 #include <SDL.h>
-#include <SDL/SDL_syswm.h>
+#include <SDL_syswm.h>
 
 #ifndef _WIN32
 #include <signal.h>
@@ -161,9 +161,9 @@ static int check_for_evdev(void)
     keycodes = XGetAtomName(info.info.x11.display, desc->names->keycodes);
     if (keycodes == NULL)
         fprintf(stderr, "could not lookup keycode name\n");
-    else if (strstart(keycodes, "evdev_", NULL))
+    else if (strstart(keycodes, "evdev", NULL))
         has_evdev = 1;
-    else if (!strstart(keycodes, "xfree86_", NULL))
+    else if (!strstart(keycodes, "xfree86", NULL))
         fprintf(stderr,
                 "unknown keycodes `%s', please report to qemu-devel@nongnu.org\n",
                 keycodes);
